/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package perceptualcolorpicker;

import javafx.scene.paint.Color;

/**
 *
 * @author Olivier
 */
/**
 * Basic color space conversion utilities, assuming two degree observer and
 * illuminant D65.
 *
 */
class ColorSpaceConvertion {

    private final double X_D65 = 0.950467;
    private final double Y_D65 = 1.0000;
    private final double Z_D65 = 1.088969;
    private final double CIE_EPSILON = 0.008856;
    private final double CIE_KAPPA = 903.3;
    private final double CIE_KAPPA_EPSILON = CIE_EPSILON * CIE_KAPPA;

    private final double b = 1.15;
    private final double g = 0.66;
    private final double c1 = 3424 / Math.pow(2, 12);
    private final double c2 = 2413 / Math.pow(2, 7);
    private final double c3 = 2392 / Math.pow(2, 7);
    private final double n = 2610 / Math.pow(2, 14);
    private final double p = 1.7 * 2523 / Math.pow(2, 5);
    private final double d = -0.56;
    private final double d0 = 1.6295499532821566 * Math.pow(10, -11);

    public ColorSpaceConvertion() {
    }

    /**
     * Conversion from CIE XYZ to sRGB as defined in the IEC 619602-1 standard
     * (http://www.colour.org/tc8-05/Docs/colorspace/61966-2-1.pdf)
     *
     * @param xyz input CIE XYZ values.
     * @param rgb output sRGB values in [0, 255] range.
     */
    public void XYZToRGB(final double[] xyz, final double[] rgb) {
        final double x = xyz[0];
        final double y = xyz[1];
        final double z = xyz[2];

        final double r_linear = +3.2406 * x - 1.5372 * y - 0.4986 * z;
        final double g_linear = -0.9689 * x + 1.8758 * y + 0.0415 * z;
        final double b_linear = +0.0557 * x - 0.2040 * y + 1.0570 * z;

        final double r = r_linear > 0.0031308
                ? 1.055 * Math.pow(r_linear, (1 / 2.4)) - 0.055
                : 12.92 * r_linear;

        final double g = g_linear > 0.0031308
                ? 1.055 * Math.pow(g_linear, (1 / 2.4)) - 0.055
                : 12.92 * g_linear;

        final double b = b_linear > 0.0031308
                ? 1.055 * Math.pow(b_linear, (1 / 2.4)) - 0.055
                : 12.92 * b_linear;

        rgb[0] = (double) (r * 255);
        rgb[1] = (double) (g * 255);
        rgb[2] = (double) (b * 255);
    }

    /**
     * Conversion from sRGB to CIE XYZ as defined in the IEC 619602-1 standard
     * (http://www.colour.org/tc8-05/Docs/colorspace/61966-2-1.pdf)
     *
     * @param rgb source sRGB values. Size of array <code>rgb</code> must be at
     * least 3. If size of array <code>rgb</code> larger than three then only
     * first 3 values are used.
     * @param xyz destinaltion CIE XYZ values. Size of array <code>xyz</code>
     * must be at least 3. If size of array <code>xyz</code> larger than three
     * then only first 3 values are used.
     */
    public void RGBToXYZ(final double[] rgb, final double[] xyz) {
        final double r = rgb[0] / 255;
        final double g = rgb[1] / 255;
        final double b = rgb[2] / 255;

        final double r_linear = r > 0.04045
                ? Math.pow((r + 0.055) / 1.055, 2.4)
                : r / 12.92;

        final double g_linear = g > 0.04045
                ? Math.pow((g + 0.055) / 1.055, 2.4)
                : g / 12.92;

        final double b_linear = b > 0.04045
                ? Math.pow((b + 0.055) / 1.055, 2.4)
                : b / 12.92;

        final double x = 0.4124 * r_linear + 0.3576 * g_linear + 0.1805 * b_linear;
        final double y = 0.2126 * r_linear + 0.7152 * g_linear + 0.0722 * b_linear;
        final double z = 0.0193 * r_linear + 0.1192 * g_linear + 0.9505 * b_linear;

        xyz[0] = (double) x;
        xyz[1] = (double) y;
        xyz[2] = (double) z;

    }

    //CIE-L*ab —> CIE-L*CH°
    public void JabToJCH(final double[] jab, final double[] jch) {
        double var_H = Math.atan2(jab[2], jab[1]); //Quadrant by signs
        if (var_H > 0) {
            var_H = (var_H / Math.PI) * 180;
        } else {
            var_H = 360 - (Math.abs(var_H) / Math.PI) * 180;
        }
        jch[0] = jab[0];
        jch[1] = (double) Math.sqrt(Math.pow(jab[1], 2) + Math.pow(jab[2], 2));
        jch[2] = (double) var_H;
    }

    //CIE-L*CH° —>CIE-L*ab//CIE-H° from 0 to 360°
    public void JCHToJab(final double[] jch, final double[] jab) {
        jab[0] = jch[0];
        jab[1] = (double) Math.cos((jch[2] * Math.PI / 180.0)) * jch[1];
        jab[2] = (double) Math.sin((jch[2] * Math.PI / 180.0)) * jch[1];

    }

    public void XYZToJab(double[] xyz, final double[] jab) {
        double[] XYZp = new double[3];
        XYZp[0] = b * xyz[0] - ((b - 1) * xyz[2]);
        XYZp[1] = g * xyz[1] - ((g - 1) * xyz[0]);
        XYZp[2] = xyz[2];

        double[] LMS = new double[3];
        LMS[0] = 0.41478972 * XYZp[0] + 0.579999 * XYZp[1] + 0.0146480 * XYZp[2];
        LMS[1] = -0.2015100 * XYZp[0] + 1.120649 * XYZp[1] + 0.0531008 * XYZp[2];
        LMS[2] = -0.0166008 * XYZp[0] + 0.264800 * XYZp[1] + 0.6684799 * XYZp[2];

        double[] LMSp = new double[3];
        for (int i = 0; i < 3; i++) {
            LMSp[i] = Math.pow((c1 + c2 * Math.pow((LMS[i] / 10000), n)) / (1 + c3 * Math.pow((LMS[i] / 10000), n)), p);
        }

        double[] Iab = new double[3];
        Iab[0] = 0.5 * LMSp[0] + 0.5 * LMSp[1];
        Iab[1] = 3.524000 * LMSp[0] - 4.066708 * LMSp[1] + 0.542708 * LMSp[2];
        Iab[2] = 0.199076 * LMSp[0] + 1.096799 * LMSp[1] - 1.295875 * LMSp[2];

        jab[0] = ((1 + d) * Iab[0] / (1 + d * Iab[0])) - d0;
        jab[1] = Iab[1];
        jab[2] = Iab[2];
    }

    public void JabToXYZ(double[] jab, double[] xyz) {

        double[] Iab = new double[3];
        Iab[0] = (jab[0] + d0) / (1 + d - d * (jab[0] + d0));
        Iab[1] = jab[1];
        Iab[2] = jab[2];

        double[] LMSp = new double[3];
        LMSp[0] = 1.0 * Iab[0] + 0.138605 * Iab[1] + 0.058047 * Iab[2];
        LMSp[1] = 1.0 * Iab[0] - 0.138605 * Iab[1] - 0.058047 * Iab[2];
        LMSp[2] = 1.0 * Iab[0] - 0.096019 * Iab[1] - 0.811892 * Iab[2];

        double[] LMS = new double[3];
        for (int i = 0; i < 3; i++) {
            LMS[i] = 10000 * Math.pow((c1 - Math.pow(LMSp[i], 1 / p)) / ((c3 * Math.pow(LMSp[i], 1 / p)) - c2), 1 / n);
        }

        double[] XYZp = new double[3];
        XYZp[0] = 1.924226 * LMS[0] - 1.004792 * LMS[1] + 0.037651 * LMS[2];
        XYZp[1] = 0.350317 * LMS[0] + 0.726481 * LMS[1] - 0.065384 * LMS[2];
        XYZp[2] = -0.090983 * LMS[0] - 0.312728 * LMS[1] + 1.522766 * LMS[2];

        xyz[0] = (XYZp[0] + (b - 1) * XYZp[2]) / b;
        xyz[1] = (XYZp[1] + (g - 1) * xyz[0]) / g;
        xyz[2] = XYZp[2];
    }

    public double[] colorToJch(Color c) {
        double[] rgb = new double[3];
        double[] xyz = new double[3];
        double[] lab = new double[3];
        double[] jch = new double[3];
        rgb[0] = (double) c.getRed() * 255;
        rgb[1] = (double) c.getGreen() * 255;
        rgb[2] = (double) c.getBlue() * 255;
        RGBToXYZ(rgb, xyz);
        XYZToJab(xyz, lab);
        JabToJCH(lab, jch);
        return jch;
    }

    public double[] colorToJab(Color c) {
        double[] rgb = new double[3];
        double[] xyz = new double[3];
        double[] jab = new double[3];
        rgb[0] = (double) c.getRed() * 255;
        rgb[1] = (double) c.getGreen() * 255;
        rgb[2] = (double) c.getBlue() * 255;
        RGBToXYZ(rgb, xyz);
        XYZToJab(xyz, jab);
        return jab;
    }
}
